#!/usr/bin/perl -w

# This is a Perl program. Get Perl from http://www.perl.com/
# This script requires the following Perl modules
# MPEG::MP3Info: http://www.perl.com/CPAN/modules/by-module/MPEG/
# File::Spec:    http://www.perl.com/CPAN/modules/by-module/File/
# Tk:            http://www.perl.com/CPAN/modules/by-module/Tk
# Tk:LEntry:     http://w4.lns.cornell.edu/~pvhp/ptk/etc/LEntry-0_00.tar.gz

# This code is distributed under the GNU General Public License (GPL). See
# http://www.opensource.org/gpl-license.html and http://www.opensource.org/.

# Please send me any modifications you make. Keep in mind that I'm likely to
# turn down obscure features to avoid including everything but the kitchen
# sink.

# The latest version is always at http://www.cs.virginia.edu/~dwc3q/code/.

# Written by: David Coppit  http://coppit.org/ <david@coppit.org>

use strict;
use Tk;
use Tk::NoteBook;
use Tk::DialogBox;
use Tk::LabFrame;

# For compiler
use Tk::Checkbutton;
use Tk::Radiobutton;
use Tk::Text;

use vars qw( %config $VERSION );

# ------------------------------------------------------------------------------

$VERSION = "0.2000";

my $dir = '';

# For some reason Tk needs the home directory. Set a default on Windows
$ENV{HOME} = 'c:\\' if !defined $ENV{HOME} &&
                      ($^O eq 'MSWin32' || $^O eq 'dos');

# Create the main window.
my $top = new MainWindow(-title => 'MP3Cleanup');

# Notebook setup
my $nb = $top->NoteBook(-ipadx => 6, -ipady => 6);

# Create the tabs for the notebook
my $main = $nb->add("main", -label => "Main", -underline => 0);
my $prefs = $nb->add("pref", -label => "Preferences", -underline => 0);

# Frame 1: Directory label, entry, and button
{
  my $dirF = $main->Frame;

  # Directory label
  $dirF->Label(-text  => "Directory: ", -anchor => 'e'
                          )->pack(-side => 'left');

  # Entry for directory
  {
    my $dirE = $dirF->Entry(-width => 20, -textvariable => \$dir);
    $dirE->bind('<Return>', sub{ DoIt($top,$dir)});
    $dirE->pack(-side => 'left',-expand => 'yes', -fill => 'x');
  }

  # Directory selection button. No good alternative yet :( ...
#  $dirF->Button(-text => "Select...", -command => sub { print $top->getOpenDir() }
#            )->pack(-side => 'left');

  $dirF->pack(-fill => 'x', -padx => '1c', -pady => 3);
}

# By default we run in recursive mode, and debug mode
$config{recursive} = 1;
$config{debug} = 1;

# Frame 2: Debug mode button, Recursive search button
{
  my $checkboxF = $main->Frame;

  # Debug mode checkbutton
  $checkboxF->Checkbutton(-text     => 'Debug mode',
                   -variable => \$config{debug},
                   -relief   => 'flat'
                )->pack(-side => 'left');

  # Recursive mode checkbutton
  $checkboxF->Checkbutton(-text     => 'Recursive',
                   -variable => \$config{recursive},
                   -relief   => 'flat'
                )->pack(-side => 'left');

  $checkboxF->pack(-anchor => 'center', -padx => '1c', -pady => '2m');
}

# Frame 3: Fix, Quit and Help buttons
{
  my $quitHelpB = $main->Frame;

  # Fix 'em button
  $quitHelpB->Button(-text => "Fix 'em!", -command => sub { DoIt($top,$dir) }
            )->pack(-side => 'left');

  $quitHelpB->Button(-text     => 'Quit',
                     -command  => [$top => 'destroy']
                )->pack(-side => 'left', -padx => '.5c');

  $quitHelpB->Button(-text     => 'Help',
                     -command  => sub{usage_help($top)}
                )->pack(-side => 'left');

  $quitHelpB->pack(-side => 'bottom', -padx => '1c', -pady => '2m');
}

# By default, we make them confirm everything.
$config{conf_case} = 1;
$config{conf_format} = 1;
$config{conf_add_tag} = 1;

# Frame 2: Preferences
{
  my $confirmF = $prefs->LabFrame(-label => 'Confirmations',
                                     -labelside => 'acrosstop');
  $confirmF->Checkbutton(
          -text     => 'Confirm change of case in file or ID3 information',
          -variable => \$config{'conf_case'},
          -relief   => 'flat'
                )->pack(-anchor => 'w');

  $confirmF->Checkbutton(
          -text     => 'Confirm change of file formatting (but not artist or song)',
          -variable => \$config{'conf_format'},
          -relief   => 'flat'
                )->pack(-anchor => 'w');

  $confirmF->Checkbutton(
          -text     => 'Confirm creation of missing ID3 artist and song from filename',
          -variable => \$config{'conf_add_tag'},
          -relief   => 'flat'
                )->pack(-anchor => 'w');

  $confirmF->pack(-expand => 'yes');

  $prefs->Button(-text     => 'Help',
                     -command  => sub{prefs_help($top)}
                )->pack(-anchor => 'center');
}

# Finally pack the notebook
$nb->pack(-expand => "yes",
     -fill => "both",
     -padx => 5, -pady => 5,
     -side => "top");

MainLoop;

# ------------------------------------------------------------------------------

use File::Spec;
use File::Find;

# ------------------------------------------------------------------------------

sub usage_help
{
  my $top = shift;

  my $helpW = $top->DialogBox(-title => 'Help',
                              -buttons => ['OK']);

  my $helpW_t = $helpW->Scrolled(qw/Text -setgrid true -width 80 -height 22 
       -wrap word -padx 1c -highlightthickness 0 -borderwidth 0 -scrollbars osoe/
    )->pack;

  $helpW_t->tag(qw/configure center -justify center -spacing1 5m -spacing3 5m/);

  $helpW_t->insert('end',<<EOF);
$0 fixes the tag information and file name for each .mp# file in
the given directory, or the current directory if none is specified.
- If the file has ID3 information, it is cleaned up.
- The filename is cleaned up.
- If the MPEG file has no ID3 information, the script attempts to extract
  it from the filename.
- If the filename is unparsable and ID3 information is available, the script
  renames the file.
- If both filename and ID3 information is available and does not agree,
  a dialog pops up so that the user can reconcile them.

Recognized file names are of the following general forms. The script handles
strange spacing fairly well:
* (Artist)-Title.mp#
* (Artist) Title.mp#
* [Artist]-Title.mp#
* [Artist] Title.mp#
* Title (Artist).mp#
* Title [Artist].mp#
* Artist - Title.mp#
* Artist _ Title.mp#
* [This-is-an-Artist]-This-is-a-Title.mp#
EOF

  $helpW->update;
  $helpW->Show;
}

# ------------------------------------------------------------------------------

sub prefs_help
{
  my $top = shift;

  my $helpW = $top->DialogBox(-title => 'Help',
                              -buttons => ['OK']);

  my $helpW_t = $helpW->Scrolled(qw/Text -setgrid true -width 80 -height 25 
       -wrap word -padx 1c -highlightthickness 0 -borderwidth 0 -scrollbars osoe/
    )->pack;

  $helpW_t->tag(qw/configure center -justify center -spacing1 5m -spacing3 5m/);

  $helpW_t->insert('end',<<EOF);
"Confirm change of case in file or ID3 information"
Uncheck this box if you want MP3Cleanup to automatically fix songs and ID3
tags where the only change is capitalization.

"Confirm change of file formatting (but not artist or song)"
Uncheck this box if you want MP3Cleanup to automatically fix file names that
have been reformatted, and the artist and song agrees with the ID3 tag.

"Confirm creation of missing ID3 artist and song from filename"
Uncheck this box if you want MP3Cleanup to automatically create ID3 tag
information from the filename, if no ID3 information already exists.

Note that if the information extracted from the file name does not match the
information from the ID3 tag, MP3Cleanup will still ask you to verify the
changes.
EOF

  $helpW->update;
  $helpW->Show;
}

#------------------------------------------------------------------------------

sub MakeDialog
{
  # Make the window
  my $changeW = $top->Toplevel(-title => 'Processing Files');

  my $numfiles = shift;

  # The rest are really references
  my $extension = shift;
  my $fileNo = shift;

  my $oldFileName = shift;
  my $oldTag = shift;

  my $suggestedFileName = shift;
  my $suggestedTag = shift;

  my $originalFileArtist = shift;
  my $originalFileSong = shift;

  my $originalTagArtist = shift;
  my $originalTagSong = shift;

  my $newFileArtist = shift;
  my $newFileSong = shift;

  my $newTagArtist = shift;
  my $newTagSong = shift;

  my $suggestedFileArtist = shift;
  my $suggestedFileSong = shift;

  my $suggestedTagArtist = shift;
  my $suggestedTagSong = shift;

  my $fileArtistChoice = shift;
  my $fileSongChoice = shift;
  my $tagArtistChoice = shift;
  my $tagSongChoice = shift;

  {
    # Frame 1: The "processing file" message
    my $proc_file = $changeW->Frame;

    $proc_file->Label(-text => "Processing file")->pack(-side => 'left');
    $proc_file->Label(-textvariable => $fileNo)->pack(-side => 'left');
    $proc_file->Label(-text => "of")->pack(-side => 'left');
    $proc_file->Label(-text => $numfiles)->pack(-side => 'left');

    $proc_file->pack;
  }

  # ----------------------------------------

  # Frame 2: Filename
  my $file_box = $changeW->LabFrame(-label     => 'File Name',
                                    -labelside => 'acrosstop');

  # Need this to work around the bug in LabFrame that doesn't allow the
  # entries to expand with the resizing of the window.
  my $file_box2 = $file_box->Frame;

  # Set up cols 2 and 4 to expand.
  $file_box2->gridColumnconfigure(2, -weight => 1);
  $file_box2->gridColumnconfigure(4, -weight => 1);

  # Do a few labels to compare the old and new file names
  {
    my $comp_files = $file_box->Frame;

    $comp_files->Label(-text => "Old:"
                   )->grid(-sticky => 'w', -row => 1, -column => 1);
    $comp_files->Label(-textvariable => $oldFileName
                   )->grid(-sticky => 'w', -row => 1, -column => 2);

    $comp_files->Label(-text => "New:"
                   )->grid(-sticky => 'w', -row => 2, -column => 1);
    $comp_files->Label(-textvariable => $suggestedFileName
                   )->grid(-sticky => 'w', -row => 2, -column => 2);

    $comp_files->pack(-side => 'bottom', -pady => 5);
  }

  # Left side is the artist
  {
    $file_box2->Label(-text => "Artist",
         )->grid(-sticky => 'we', -columnspan => 2, -row => 1, -column => 1);

    my $temp_entry = $file_box2->Entry(-textvariable => $originalFileArtist,
                                       -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 2, -column => 2);

    $file_box2->Radiobutton(
              -text     => "File:",
              -variable => $fileArtistChoice,
              -command  => sub {
                     $$suggestedFileArtist = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},
              -relief   => 'flat',
              -value    => 'File',
             )->grid(-sticky => 'w', -row => 2, -column => 1);
  }

  {
    my $temp_entry = $file_box2->Entry(-textvariable => $originalTagArtist,
                                       -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 3, -column => 2);

    $file_box2->Radiobutton(
              -text     => "ID3:",
              -variable => $fileArtistChoice,
              -command  => sub {
                     $$suggestedFileArtist = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},
              -relief   => 'flat',
              -value    => 'Tag',
             )->grid(-sticky => 'w', -row => 3, -column => 1);
  }

  {
    my $temp_entry = $file_box2->Entry(-width        => 40,
                                       -textvariable => $newFileArtist);
    $temp_entry->bind('<KeyPress>', sub{
                  if ($$fileArtistChoice eq 'Guess')
                  {
                     $$suggestedFileArtist = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);
                  }});
    $temp_entry->grid(-sticky => 'we', -row => 4, -column => 2);

    $file_box2->Radiobutton(
              -text     => "Guess:",
              -variable => $fileArtistChoice,
              -command  => sub {
                     $$suggestedFileArtist = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},
              -relief   => 'flat',
              -value    => 'Guess',
             )->grid(-sticky => 'w', -row => 4, -column => 1);
  }


  # Right side is the song

  {
    $file_box2->Label(-text => "Song",
         )->grid(-sticky => 'we', -columnspan => 2, -row => 1, -column => 3);

    my $temp_entry = $file_box2->Entry(-textvariable => $originalFileSong,
                                       -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 2, -column => 4);

    $file_box2->Radiobutton(
              -text     => "File:",
              -variable => $fileSongChoice,
              -command  => sub {
                     $$suggestedFileSong = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},
              -relief   => 'flat',
              -value    => 'File',
             )->grid(-sticky => 'w', -row => 2, -column => 3);
  }

  {
    my $temp_entry = $file_box2->Entry(-textvariable => $originalTagSong,
                                       -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 3, -column => 4);

    $file_box2->Radiobutton(
              -text     => "ID3:",
              -variable => $fileSongChoice,
              -command  => sub {
                     $$suggestedFileSong = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},

              -relief   => 'flat',
              -value    => 'Tag',
             )->grid(-sticky => 'w', -row => 3, -column => 3);
  }

  {
    my $temp_entry = $file_box2->Entry(-width        => 40,
                                       -textvariable => $newFileSong);
    $temp_entry->bind('<KeyPress>', sub{
                  if ($$fileSongChoice eq 'Guess')
                  {
                    $$suggestedFileSong = $temp_entry->get;
                    $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                      $$suggestedFileSong,
                                                      $$extension);
                  }});
    $temp_entry->grid(-sticky => 'we', -row => 4, -column => 4);

    $file_box2->Radiobutton(
              -text     => "Guess:",
              -variable => $fileSongChoice,
              -command  => sub {
                     $$suggestedFileSong = $temp_entry->get;
                     $$suggestedFileName = get_new_name($$suggestedFileArtist,
                                                       $$suggestedFileSong,
                                                       $$extension);},
              -relief   => 'flat',
              -value    => 'Guess',
             )->grid(-sticky => 'w', -row => 4, -column => 3);
  }

  $file_box2->pack(-fill => 'x');
  $file_box->pack(-side => 'top', -fill => 'x', -padx => 5, -pady => 5);

  # ----------------------------------------------------------------------

  # Frame 3: ID3
  my $tag_box = $changeW->LabFrame(-label => 'ID3',
                                   -labelside => 'acrosstop');

  # Need this to work around the bug in LabFrame that doesn't allow the
  # entries to expand with the resizing of the window.
  my $tag_box2 = $tag_box->Frame;

  # Set up cols 2 and 4 to expand.
  $tag_box2->gridColumnconfigure(2, -weight => 1);
  $tag_box2->gridColumnconfigure(4, -weight => 1);

  # Do a few labels to compare the old and new tags
  {
    my $comp_tags = $tag_box->Frame;

    $comp_tags->Label(-text => "Old:"
                   )->grid(-sticky => 'w', -row    => 1,   -column     => 1);
    $comp_tags->Label(-textvariable => $oldTag
                   )->grid(-sticky => 'w', -row    => 1,   -column     => 2);

    $comp_tags->Label(-text => "New:"
                   )->grid(-sticky => 'w', -row => 2, -column => 1);
    $comp_tags->Label(-textvariable => $suggestedTag
                   )->grid(-sticky => 'w', -row => 2, -column => 2);

    $comp_tags->pack(-side => 'bottom', -pady => 5);
  }


  # Left side is the artist
  {
    $tag_box2->Label(-text => "Artist",
         )->grid(-sticky => 'we', -columnspan => 2, -row => 1, -column => 1);

    my $temp_entry = $tag_box2->Entry(-textvariable => $originalFileArtist,
                                      -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 2, -column => 2);

    $tag_box2->Radiobutton(
              -text     => "File:",
              -variable => $tagArtistChoice,
              -command  => sub {
                $$suggestedTagArtist = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'File',
             )->grid(-sticky => 'w', -row => 2, -column => 1);
  }

  {
    my $temp_entry = $tag_box2->Entry(-textvariable => $originalTagArtist,
                                      -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 3, -column => 2);

    $tag_box2->Radiobutton(
              -text     => "ID3:",
              -variable => $tagArtistChoice,
              -command  => sub {
                $$suggestedTagArtist = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'Tag',
             )->grid(-sticky => 'w', -row => 3, -column => 1);
  }

  {
    my $temp_entry = $tag_box2->LEntry(-maxwidth     => 30,
                                       -width        => 40,
                                       -textvariable => $newTagArtist);
    $temp_entry->bind('<KeyPress>', sub{
             if ($$tagArtistChoice eq 'Guess')
             {
               $$suggestedTagArtist = $temp_entry->get;
               $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";
             }});
    $temp_entry->grid(-sticky => 'we', -row => 4, -column => 2);

    $tag_box2->Radiobutton(
              -text     => "Guess:",
              -variable => $tagArtistChoice,
              -command  => sub {
                $$suggestedTagArtist = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'Guess',
             )->grid(-sticky => 'w', -row => 4, -column => 1);
  }


  # Right side is the song
  {
    $tag_box2->Label(-text => "Song",
         )->grid(-sticky => 'we', -columnspan => 2, -row => 1, -column => 3);

    my $temp_entry = $tag_box2->Entry(-textvariable => $originalFileSong,
                                      -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 2, -column => 4);

    $tag_box2->Radiobutton(
              -text     => "File:",
              -variable => $tagSongChoice,
              -command  => sub {
                $$suggestedTagSong = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'File',
             )->grid(-sticky => 'w', -row => 2, -column => 3);
  }

  {
    my $temp_entry = $tag_box2->Entry(-textvariable => $originalTagSong,
                                      -state        => 'disabled');
    $temp_entry->grid(-sticky => 'we', -row => 3, -column => 4);

    $tag_box2->Radiobutton(
              -text     => "ID3:",
              -variable => $tagSongChoice,
              -command  => sub {
                $$suggestedTagSong = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'Tag',
             )->grid(-sticky => 'w', -row => 3, -column => 3);
  }

  {
    my $temp_entry = $tag_box2->LEntry(-maxwidth     => 30,
                                       -width        => 40,
                                       -textvariable => $newTagSong);
    $temp_entry->bind('<KeyPress>', sub{
             if ($$tagSongChoice eq 'Guess')
             {
               $$suggestedTagSong = $temp_entry->get;
               $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";
             }});
    $temp_entry->grid(-sticky => 'we', -row => 4, -column => 4);

    $tag_box2->Radiobutton(
              -text     => "Guess:",
              -variable => $tagSongChoice,
              -command  => sub {
                $$suggestedTagSong = $temp_entry->get;
                $$suggestedTag = "$$suggestedTagArtist :: $$suggestedTagSong";},
              -relief   => 'flat',
              -value    => 'Guess',
             )->grid(-sticky => 'w', -row => 4, -column => 3);
  }

  $tag_box2->pack(-fill => 'x');

  $tag_box->pack(-side => 'top', -fill => 'x', -padx => 5, -pady => 5);

  $changeW->withdraw;
  return $changeW;
}

#------------------------------------------------------------------------------

sub ShowDialog
{
  my $changeW = shift;

  my $oldFileName = shift;

  # Really a ref
  my $suggestedFileName = shift;
  my $oldTag = shift;

  # Really a ref
  my $suggestedTag = shift;

  my $originalFileArtist = shift;
  my $originalFileSong = shift;

  my $originalTagArtist = shift;
  my $originalTagSong = shift;

  # These are refs
  my $suggestedFileArtist = shift;
  my $suggestedFileSong = shift;

  my $suggestedTagArtist = shift;
  my $suggestedTagSong = shift;


  my $file_frame = ($changeW->children)[1];
  my $tag_frame = ($changeW->children)[2];

  $file_frame->packForget;
  $tag_frame->packForget;

  # ----------------- FIGURE OUT WHAT TO SHOW IN THE DIALOG -----------------

  my $show_file = 1;
  my $show_tag = 1;
  my $fileChoice = 'Skip';
  my $tagChoice = 'Skip';

  # Abort if nothing is changed and tag matches filename
  return ('Skip','Skip')
    if ($oldFileName eq $$suggestedFileName) && ($oldTag eq $$suggestedTag) &&
       matches($originalFileArtist,$originalTagArtist) &&
       matches($originalFileSong,$originalTagSong);

  # Don't show if nothing is changed in filename or tag
  $show_file = 0, $fileChoice = 'Skip' if $oldFileName eq $$suggestedFileName;
  $show_tag = 0, $tagChoice = 'Skip' if $oldTag eq $$suggestedTag;

  # Do it automatically if only case is changed, and prefs say to skip this
  $show_file = 0, $fileChoice = 'Change'
    if !$config{conf_case} && (lc $oldFileName eq lc $$suggestedFileName) &&
                              ($oldFileName ne $$suggestedFileName);
  $show_tag = 0, $tagChoice = 'Change'
    if !$config{conf_case} && (lc $oldTag eq lc $$suggestedTag) &&
                              ($oldTag ne $$suggestedTag);

  # Do it automatically if artist and song are the same but formatting is 
  # different, and prefs say to skip this
  $show_file = 0, $fileChoice = 'Change'
    if !$config{conf_format} &&
      (($config{conf_case} && ($originalFileArtist eq $$suggestedFileArtist) &&
                              ($originalFileSong eq $$suggestedFileSong)) ||
       ((!$config{conf_case} &&
             (lc $originalFileArtist eq lc $$suggestedFileArtist) &&
             (lc $originalFileSong eq lc $$suggestedFileSong))));
  $show_tag = 0, $tagChoice = 'Change'
    if !$config{conf_format} &&
      (($config{conf_case} && ($originalTagArtist eq $$suggestedTagArtist) &&
                              ($originalTagSong eq $$suggestedTagSong)) ||
       ((!$config{conf_case} &&
             (lc $originalTagArtist eq lc $$suggestedTagArtist) &&
             (lc $originalTagSong eq lc $$suggestedTagSong))));

  # Don't confirm the creation of ID3 from filename
  $show_tag = 0, $tagChoice = 'Change'
    if !$config{conf_add_tag} &&
      ($originalTagArtist eq '') && ($originalTagSong eq '') &&
      ($$suggestedTagArtist ne '') && ($$suggestedTagSong ne '');

  # However, show both if the tag and file info don't match
  $show_tag = 1, $show_file = 1
    if !matches($originalFileArtist,$originalTagArtist) ||
       !matches($originalFileSong,$originalTagSong);

  return ($fileChoice,$tagChoice) if !$show_file && !$show_tag;

  # --------------------------------------------------------------------------

  # Set a grab and claim the focus
  my $old_focus = $changeW->focusSave;
  my $old_grab  = $changeW->grabSave;

  $changeW->grab;
  $changeW->focus;
  $changeW->deiconify;

  $file_frame->pack(-side => 'top', -fill => 'x', -padx => 5, -pady => 5)
    if $show_file;
  $tag_frame->pack(-side => 'top', -fill => 'x', -padx => 5, -pady => 5)
    if $show_tag;

  my $okCancelQuit;

  # Frame 4: Ok, Cancel, Quit buttons
  my $buttonF = $changeW->Frame;
  {

    $buttonF->Button(-text    => "Change",
                     -command => sub{$okCancelQuit = 'Change'}
                   )->pack(-side => 'left');
    $buttonF->Button(-text    => "Skip",
                     -command => sub{$okCancelQuit = 'Skip'}
                   )->pack(-side => 'left');
    $buttonF->Button(-text    => "Stop",
              -command => sub {$okCancelQuit = 
                                'Stop'}
                   )->pack(-side => 'left');

    $buttonF->pack(-side => 'bottom');
  }

  $changeW->update;

  {
    my $PosX = int(($top->screenwidth - $changeW->reqwidth) / 2);
    my $PosY = int(($top->screenheight - $changeW->reqheight) / 2);
    $changeW->geometry($changeW->reqwidth.'x'.$changeW->reqheight.'+'.$PosX.'+'.$PosY);
  }

  $changeW->waitVariable(\$okCancelQuit);

  $buttonF->destroy;
  $changeW->withdraw;

  # Restore the grab and focus
  $changeW->grabRelease;
  &$old_focus;
  &$old_grab;

  # --------------------------------------------------------------------------

  # First take care of the "Skip" and "Stop" selections.
  return ('Stop','Stop') if $okCancelQuit eq 'Stop';

  if ($okCancelQuit eq 'Skip')
  {
    return ('Skip','Skip') if $show_file == 1 && $show_tag == 1;
    return ($fileChoice,'Skip') if $show_file == 0 && $show_tag == 1;
    return ('Skip',$tagChoice) if $show_file == 1 && $show_tag == 0;
  }

  if ($okCancelQuit eq 'Change')
  {
    $fileChoice = 'Change',$tagChoice = 'Change'
      if $show_file == 1 && $show_tag == 1;
    $fileChoice = 'Change' if $show_file == 1 && $show_tag == 0;
    $tagChoice = 'Change' if $show_file == 0 && $show_tag == 1;
  }

  # Since the user may have changed the info into the way it was at the
  # beginning, we have to check it one more time

  $fileChoice = 'Skip' if $oldFileName eq $$suggestedFileName;
  $tagChoice = 'Skip' if $oldTag eq $$suggestedTag;

  return ($fileChoice,$tagChoice);
}

#------------------------------------------------------------------------------

use MPEG::MP3Info;
use File::Basename;
use File::Spec::Functions;
use File::Find;
use Tk::LEntry;
use strict;

sub DoIt
{
  my $top = shift;
  my $directory = shift;

  unless (-d $directory)
  {
    my $dialog = $top->DialogBox(-title => 'Error', -buttons => ['OK']);
    $dialog->Label(-bitmap => 'error')->pack(-side => 'left');
    my $message = "$directory is not a directory.\n";
    $dialog->Label(-text => $message,
                      -wraplength => '6i')->pack(-side => 'left');

    $dialog->update;
    $dialog->Show;
    return;
  }

  # Get the songs
  my @files = get_files($directory);

  # Have to predeclare these for the dialog box
  my ($extension,$fileNo,$oldFileName,$oldTag,$suggestedFileName,$suggestedTag,
      $originalFileArtist,$originalFileSong,$originalTagArtist,
      $originalTagSong,$newFileArtist,$newFileSong,$newTagArtist,$newTagSong,
      $suggestedFileArtist,$suggestedFileSong,$suggestedTagArtist,
      $suggestedTagSong,$fileArtistChoice,$fileSongChoice,$tagArtistChoice,
      $tagSongChoice);

  my $changeW = 
    MakeDialog($#files,\$extension,
             \$fileNo,\$oldFileName,\$oldTag,\$suggestedFileName,
             \$suggestedTag,\$originalFileArtist,\$originalFileSong,
             \$originalTagArtist,\$originalTagSong,\$newFileArtist,
             \$newFileSong,\$newTagArtist,\$newTagSong,\$suggestedFileArtist,
             \$suggestedFileSong,\$suggestedTagArtist,\$suggestedTagSong,
             \$fileArtistChoice,\$fileSongChoice,\$tagArtistChoice,
             \$tagSongChoice);

  # Now make the log, unless we already did
  {
    my $label;

    if (defined $top->{MYLABEL})
    {
      $label = $top->{MYLABEL};
    }
    else
    {
      $label = $top->Label;
      $top->{MYLABEL} = $label;
    }

    my $text = "Change Log";
    $text .= " (Debug mode -- no changes made)" if $config{debug};
    $label->configure(-text => $text);
    $label->pack(-side => 'top');
  }

  my $log;
  if (defined $top->{MYLOG})
  {
    $log = $top->{MYLOG};
    $log->delete('1.0','end');
  }
  else
  {
    $log = $top->Scrolled(qw/Text -setgrid true -width 80 -height 25 
       -wrap word -padx 1c -highlightthickness 0 -borderwidth 0
       -scrollbars osoe/
      )->pack;
    $top->{MYLOG} = $log;
  }

  # Process each file
  $fileNo = 0;
  foreach my $fullpath (@files)
  {
    $log->insert('end',"Processing file:\n  $fullpath\n");
    $log->yview('end');
    $log->update;

    $fileNo++;
    my ($fileinfo, $taginfo, $filename, $directory);

    ($filename, $directory, $extension) = fileparse($fullpath, '(?i)\.mp\d');

    $oldFileName = "$filename$extension";

    $extension = lc($extension);
    $filename .= $extension;

    # ---------------- FIRST STEP: GET INFO FROM FILE AND TAG -----------------

    # Get the original artist and title
    ($originalFileArtist,$originalFileSong) = get_info_from_file($filename);
    ($originalTagArtist,$originalTagSong) = get_info_from_tag($fullpath);

    $oldTag = "$originalTagArtist :: $originalTagSong";

    # ---------------- SECOND STEP: CLEAN UP FILE AND TAG INFO ----------------

    # Clean up the artist and title
    my $cleanFileArtist = make_artist_pretty($originalFileArtist);
    my $cleanFileSong = make_title_pretty($originalFileSong);
    my $cleanTagArtist = make_artist_pretty($originalTagArtist);
    my $cleanTagSong = make_title_pretty($originalTagSong);

    $cleanTagArtist = shorten_tag($cleanTagArtist)
      if length $cleanTagArtist > 30;
    $cleanTagSong = shorten_tag($cleanTagSong)
      if length $cleanTagSong > 30;

    # ------- THIRD STEP: TRY TO AUTOMATICALLY RESOLVE ANY CONFLICTS ----------

    # By default, we only pretty up the names
    $newFileArtist = $cleanFileArtist;
    $newFileSong = $cleanFileSong;
    $newTagArtist = $cleanTagArtist;
    $newTagSong = $cleanTagSong;

    $newFileArtist = $newTagArtist
      if ($newFileArtist eq '') && ($newTagArtist ne '');
    $newTagArtist = $newFileArtist
      if ($newTagArtist eq '') && ($newFileArtist ne '');
    $newFileSong = $newTagSong
      if ($newFileSong eq '') && ($newTagSong ne '');
    $newTagSong = $newFileSong
      if ($newTagSong eq '') && ($newFileSong ne '');

    # --------- FOURTH STEP: SET THE STATE OF THE DIALOG BOX AND LET ----------
    # ---------------------- THE USER MAKE A DECISION -------------------------

    $fileArtistChoice = 'Guess';
    $fileSongChoice = 'Guess';
    $suggestedFileArtist = $newFileArtist;
    $suggestedFileSong = $newFileSong;
    $suggestedFileName = get_new_name($suggestedFileArtist,
                                      $suggestedFileSong,
                                      $extension);

    $tagArtistChoice = 'Guess';
    $tagSongChoice = 'Guess';
    $suggestedTagArtist = $newTagArtist;
    $suggestedTagSong = $newTagSong;
    $suggestedTag = "$suggestedTagArtist :: $suggestedTagSong";

    my ($fileChoice,$tagChoice) = 
      ShowDialog($changeW,$oldFileName,\$suggestedFileName,
                 "$originalTagArtist :: $originalTagSong", \$suggestedTag,
                 $originalFileArtist,$originalFileSong,$originalTagArtist,
                 $originalTagSong,\$suggestedFileArtist,\$suggestedFileSong,
                 \$suggestedTagArtist,\$suggestedTagSong);

    if ($config{debug})
    {
      $log->insert('end',"File decision: $fileChoice, Tag decision: $tagChoice\n");
      $log->yview('end');
      $log->update;
    }

    if ($fileChoice eq 'Stop')
    {
      $changeW->destroy;
      return;
    }

    # --------------- FIFTH STEP: RENAME OR RETAG THE FILE --------------------

    if ($tagChoice eq 'Change')
    {
      if ($config{debug})
      {
        $log->insert('end',"Setting tag:\n  Artist: $suggestedTagArtist\n  Song: $suggestedTagSong\n");
        $log->yview('end');
        $log->update;
      }

      my $mp3_tag = get_mp3tag($fullpath);
      $mp3_tag->{ARTIST} = $suggestedTagArtist;
      $mp3_tag->{TITLE} = $suggestedTagSong;
      $mp3_tag->{GENRE} = 'Sound Clip' unless defined $mp3_tag->{GENRE};
      $mp3_tag->{ALBUM} = '' unless defined $mp3_tag->{ALBUM};
      $mp3_tag->{YEAR} = '' unless defined $mp3_tag->{YEAR};
      $mp3_tag->{COMMENT} = '' unless defined $mp3_tag->{COMMENT};
      $mp3_tag->{TRACKNUM} = 255 unless defined $mp3_tag->{TRACKNUM};
      set_mp3tag($fullpath, $mp3_tag) unless $config{debug};
    }

    if ($fileChoice eq 'Change')
    {
      my $newname = get_new_name($suggestedFileArtist,$suggestedFileSong,
                                 $extension);

      if ($config{debug})
      {
        $log->insert('end',"Setting filename:\n  Name: $newname\n");
        $log->yview('end');
        $log->update;
      }

      unless ($config{debug})
      {
        if (rename($fullpath, catfile($directory, $newname)))
        {
          $fullpath = catfile($directory,$newname);
        }
        else
        {
          # In Windows, we get this wierd message when renaming a file that
          # just has a change in the case. It seem to rename okay, so we'll
          # just ignore it.
          unless (($^O eq 'MSWin32' || $^O eq 'dos') &&
                  ($! eq 'No such file or directory'))
          {
            my $dialog = $top->DialogBox(-title => 'Error', -buttons => ['OK']);
            $dialog->Label(-bitmap => 'error')->pack(-side => 'left');
            my $message = "Can't rename '$filename' to '$newname': $!\n";
            $dialog->Label(-text => $message,
                              -wraplength => '6i')->pack(-side => 'left');

            $dialog->update;
            $dialog->Show;
          }
        }
      }
    }
  }

  $changeW->destroy;
  return;
}

# -----------------------------------------

sub get_info_from_file
{
  my $filename = shift;

  my $fileinfo = process_filename($filename);
  return ($fileinfo->{ARTIST},$fileinfo->{TITLE});
}

# -----------------------------------------

sub get_info_from_tag
{
  my $filename = shift;

  my $taginfo = process_tag($filename);
  return ($taginfo->{ARTIST},$taginfo->{TITLE});
}

# -----------------------------------------

sub confirm_file_overwrite
{
  my $filename = shift;

  my $dialog = $top->DialogBox(-title => 'Warning',
                                      -buttons => ['Yes','No']);
  $dialog->Label(-bitmap => 'warning', -padx => '.5c',
                    -wraplength => '6i')->pack(-side => 'left');

  my $message = "The file ".catfile($filename).
     " already exists. Would you like to overwrite it?";
  $dialog->Label(-text => $message, -padx => '.5c',
                      -wraplength => '6i')->pack(-side => 'left');

  $dialog->update;
  return $dialog->Show;
}

#------------------------------------------------------------------------------
    
sub make_artist_pretty
{
  my $artist = shift;

  # Change _ to space
  $artist =~ s/_/ /g;

  # Change - to space if there are no spaces
  $artist =~ s/-/ /g unless $artist =~ / /;

  # Change %20 to space
  $artist =~ s/%20/ /g;

  # Change extra spaces to 1 space
  $artist =~ s/  */ /g;

  # Change extra - --- -- to 1 -
  $artist =~ s/-[\s-]*-/-/g;

  # Make "TheArtist" become "The Artist" except MacDonald or McDonald
  $artist =~ s/([a-z])([A-Z])/$1 $2/g;
  $artist =~ s/(Ma?c) ([A-Z])/$1$2/g;

  # Upper case all words
  $artist =~ s/(^| )(\w)/$1\u$2/g;

  # Remove preceding spaces
  $artist =~ s/^ *//;

  return $artist;
}

#------------------------------------------------------------------------------

sub make_title_pretty
{
  my $title = shift;

  # Change _ to space
  $title =~ s/_/ /g;

  # Change - to space if there are no spaces
  $title =~ s/-/ /g unless $title =~ / /;

  # Change %20 to space
  $title =~ s/%20/ /g;

  # Change extra spaces to 1 space
  $title =~ s/  */ /g;

  # Change extra - --- -- to 1 -
  $title =~ s/-[\s-]*-/-/g;

  # Make "TheSong" become "The Song", except McDonald and MacDonald
  $title =~ s/([a-z])([A-Z])/$1 $2/g;
  $title =~ s/(Ma?c) ([A-Z])/$1$2/g;

  # Make "Words(blah)" look like "Words (blah)", and "Words#1" look like
  # "Words #1"
  $title =~ s/(\S)([\(\[#])/$1 $2/g;
 
  # Make "Words&Stuff" look like "Words & Stuff"
  $title =~ s/([^\s])&([^\s])/$1 & $2/g;

  # Make "Words(blah)" look like "Words (blah)", and "Words#1" look like
  # "Words #1"
  $title =~ s/(\S)([\(\[#])/$1 $2/g;
 
  # Make "Words32" look like "Words 32"
  $title =~ s/([^\d\s\-#])(\d+)/$1 $2/g;

  # Make "(blah)" look like "(Blah)"
  $title =~ s/\((.*)\)/\(\u$1\)/;

  # Title case it all
  $title =~ s/(^|\s)(\w)/$1\U$2/g;

  # Uppercase roman numerals
  $title =~ s/\b(ii+)\b/\U$1/ig;

  # Remove preceding spaces
  $title =~ s/^ *//;

  return $title;
}

#------------------------------------------------------------------------------

sub shorten_tag
{
  my $taginfo = shift;

  return if length $taginfo <= 30;

  # First try to remove some spaces
  my $nospaces = $taginfo;
  $nospaces =~ s/ //g;
  if (length $nospaces <= 30)
  {
    $taginfo =~ s/(.*) /$1/ while length $taginfo > 30;
  }
  else
  {
    $taginfo = substr($taginfo,0,27)."...";
  }

  return $taginfo;
}

#------------------------------------------------------------------------------

sub process_tag
{
  my $fullpath = shift;

  my $taginfo = get_mp3tag($fullpath);

  if (!defined $taginfo)
  {
    return {'ARTIST' => '','TITLE' => '','ALBUM' => '',
            'GENRE' => 'Sound Clip', 'COMMENT' => '', 'YEAR' => ''};
  }
  else
  {
    return $taginfo;
  }
}

#------------------------------------------------------------------------------

sub process_filename
{
  my $filename = shift;

  # First strip any track number from the front, or after a -. Try to keep
  # valid ones by weeding out any greater than 15.
  $filename =~ s/^\s*\d\d\s+-?\s*//
    if ($filename =~ /^\s*(\d\d)\s+/) && ($1 < 15);
  $filename =~ s/-\s+\d\d\s+/-/
    if ($filename =~ /-\s+(\d\d)\s+/) && ($1 < 15);

  # Collapse -- -, --, - -, to -
  $filename =~ s/-[-\s]+-/-/g;

  my ($artist, $song);

  if (
      # Check "Artist - Song (stuff)"
      ($filename =~ /^([^\(\)\[\]-]+?)-(.+\(.+\).+?).mp\d$/i) ||
      # Check "(Artist)-Song"
      ($filename =~ /^[\s_]*\(([^\)]+?)\)[\s_]*-+(.+?).mp\d$/i) ||
      # Check "(Artist) Song"
      ($filename =~ /^[\s_]*\(([^\)]+?)\)(.+?).mp\d$/i) ||
      # Check "[Artist] - Song"
      ($filename =~ /^[\s_]*\[([^\]]+?)\][\s_]*-(.+?).mp\d$/i) ||
      # Check "[Artist] Song"
      ($filename =~ /^[\s_]*\[([^\)]+?)\](.+?).mp\d$/i) ||
      # Check "Artist-O - Song-O"
      ($filename =~ /^((?:[^-]|\S-\S)+?)\s+-\s+((?:[^-]|\S-\S)+?).mp\d$/i) ||
      # Check "Artist - Song", "Artist- Song", "Artist--Song"
      ($filename =~ /^([^-]+?)\s*-+\s*([^-\[]+?).mp\d$/i) ||
      # Check "Artist _ Song", "Artist_ Song", "Artist__Song"
      ($filename =~ /^([^_]+?)\s*_\s*([^_]+?).mp\d$/i))
  {
    $artist = $1;
    $song = $2;
  }
  elsif (
         # Check "Song-O - (Artist-O)"
         ($filename =~ /^((?:[^\(\)-]|\S-\S)+?)\s+-\s+\(([^\)]+?)\)[\s_]*.mp\d$/i) ||
         # Check "Song - (Artist)"
         ($filename =~ /^([^\(\)\[\]-]+?)-[\s_]*\(([^\)]+?)\)[\s_]*.mp\d$/i) ||
         # Check "Song - [Artist]"
         ($filename =~ /^([^\[\]-]+?)-[\s_]*\[([^\]]+?)\][\s_]*.mp\d$/i) ||
         # Check "Song (Artist)"
         ($filename =~ /^([^\(\)\[\]-]+?)[\s_]*\(([^\)]+?)\)[\s_]*.mp\d$/i) ||
         # Check "Song [Artist]"
         ($filename =~ /^([^\(\)\[\]-]+?)[\s_]*\[([^\]]+?)\][\s_]*.mp\d$/i))
  {
    $artist = $2;
    $song = $1;
  }
  else
  {
    return {'ARTIST' => '','TITLE' => ''};
  }

  # Strip any leading and trailing whitespace
  $artist =~ s/^[\s_]*(.*?)[\s_]*$/$1/;
  $song =~ s/^[\s_]*(.*?)[\s_]*$/$1/;

  return {'ARTIST' => $artist,'TITLE' => $song};
}

#------------------------------------------------------------------------------

sub get_new_name
{
  my($artist, $title, $extension) = @_;
  my $newname;

  $extension = lc($extension);

  if ($^O eq 'MacOS')
  {
    $artist = substr($artist,0,13);
    $title = substr($title,0,13);
  }

  $newname = '';

  $newname = "[$artist]" if $artist ne '';
  $newname .= " " if $artist ne '' && $title ne '';
  $newname .= $title if $title ne '';
  $newname .= $extension;

  $newname =~ tr/:\?"/-/;

  return $newname;
}

#------------------------------------------------------------------------------

sub matches
{
  my($fileinfo, $taginfo) = @_;

  return 1 if $fileinfo eq $taginfo;

  # -----------------------------

  # It matches if any wierd character got turned into -'s for the filename
  my $tagNoWierdChars = $taginfo;
  $tagNoWierdChars =~  tr/:\?"/-/;

  return 1 if $fileinfo eq $tagNoWierdChars;

  # Make taginfo tagNoWierdChars for the rest just in case this happened in
  # addition to some other change.
  $taginfo = $tagNoWierdChars;

  # -----------------------------

  # If they want to ignore case the match might occur. Go ahead and keep this
  # change in case it happens in conjunction with some other change
  $fileinfo = lc $fileinfo if !$config{conf_case};
  $taginfo = lc $taginfo if !$config{conf_case};

  return 1 if $fileinfo eq $taginfo;

  # Check if they removed spaces to fit the info into 30 characters
  my $filenospaces = $fileinfo;
  $filenospaces =~ s/ //g;

  my $tagnospaces = $taginfo;
  $tagnospaces =~ s/ //g;

  return 1 if $filenospaces eq $tagnospaces;

  # -----------------------------

  # Check if they did .. to shorten the tag info
  if ($taginfo =~ /^(.*)\.\.+(.*)$/)
  {
    return 1 if $fileinfo =~ /^$1.*$2$/;
  }

  # -----------------------------

  # Check if they just truncated the info
#  return 1 if substr($fileinfo,0,30) eq $taginfo;

  # -----------------------------

  return 0;
}

#------------------------------------------------------------------------------

sub get_files
{
  my $directory = shift;
  my @files;

  # Make sure there's a trailing slash for File::Find
  $directory =~ s#\\#/#g;
  $directory .= '/' unless $directory =~ /\/$/;

  # Do a recursive search for files if they want it.
  if ($config{recursive} == 1)
  {
    # Gets all mp3 files in current directory and descendents. Puts them in
    # @files
    find(sub {
        push @files,"$File::Find::name"
          if $File::Find::name =~ /\.mp\d+$/i
      },$directory);
  }
  # Otherwise just read
  else
  {
    opendir DIR, $directory or die "Can't open $directory\n";
    my @dirfiles = readdir DIR;
    closedir DIR;
    
    foreach my $file (@dirfiles)
    {
      push @files,catfile($directory,$file) if !-d $file && $file =~ /\.mp\d+$/i;
    }
  }

  # Flip all slashes forward, for prettiness
  grep {s#\\#/#g} @files if $^O eq 'Win32';

  return sort @files;
}

